/*  ***** BEGIN LICENSE BLOCK *****
 *  Version: RPPL 1.0/GPL 2.0/LGPL 2.1/MPL 1.1/CDDL 1.0
 *  
 *  
 *  The contents of this file are subject to the Roland Pihlakas Public
 *  License Version 1.0 (the "License"); you may not use this file except
 *  in compliance with the License. You may obtain a copy of the License
 *  at http://roland.pri.ee/RPPL/
 *  
 *  Software distributed under the License is distributed on an "AS IS"
 *  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 *  License for the specific language governing rights and limitations
 *  under the License.
 *  
 *  The Original Code is: Modelling of Natural Thinking.
 *  
 *  The Initial Developer of the Original Code is: Roland Pihlakas.
 *  
 *  Portions created by: _____________________________ 
 *  are Copyright (C): _____________________________. 
 *  All Rights Reserved.
 *  Contributor(s): ______________________________________.
 *  
 *  Alternatively, the contents of this file may be used under the terms of 
 *  either the GNU General Public License Version 2 or later (the "GPL"), or 
 *  the GNU Lesser General Public License Version 2.1 or later (the "LGPL"), or 
 *  the Mozilla Public Licence Version 1.1 or later (the "MPL"), or the Common 
 *  Development and Distribution License Version 1.0 or later (the "CDDL"), in 
 *  which case the provisions of the MPL, the GPL, the LGPL or the CDDL are 
 *  applicable instead of those above. If you wish to allow use of your version 
 *  of this file only under the terms of the of either the MPL, the GPL, 
 *  the LGPL or the CDDL, and not to allow others to use your version of this 
 *  file under the RPPL, indicate your decision by deleting the provisions 
 *  above and replace them with the notice and other provisions required by 
 *  the MPL, the GPL, the LGPL or the CDDL. If you do not delete the provisions 
 *  above, a recipient may use your version of this file under the terms of any 
 *  one of the RPPL, the MPL, the GPL, the LGPL or the CDDL.
 *
 *  ***** END LICENSE BLOCK ***** */

//######################################################################################################
//
//######################################################################################################




#ifndef STATUS_VECTOR_H
#define STATUS_VECTOR_H



/// MS compatible compilers support #pragma once
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif



///#######################################################################
///
///#######################################################################


#include <stdarg.h>



#include <map>




#include "../macros.h"
#include "../definitions.h"
#include "../datastructures/datatypes.h"
#include "../datastructures/linked_list.h"
#include "../tasand1/tasand1.h"
///#include "../tasand1/cell_value.h"
#include "../tasand1/cvop.h"







namespace ai
{

	using namespace definitions;
	using namespace datatypes;
	using namespace linked_list;

	using namespace std;			///map object






///#######################################################################
///
///#######################################################################

template
<
	typename val_T
>
class my_ptr_list {			

protected:

	sllc<ptrs_list_elem<val_T>/*, DSS*/>		pointers_a;

	///##################################################################

public:
	my_ptr_list(val_T *val_ptr1, ...)
	{	
		val_T *ptr = val_ptr1;


		va_list marker;
		va_start(marker, val_ptr1);   
		while (ptr != NULL)			   ///-1 is terminator
		{
			pointers_a.additem(ptr);
			
			ptr = va_arg(marker, val_T *);
		};
		va_end(marker);		
	};

	///##################################################################
	
	V operator = (IN sllc<T_list_elem<val_T>/*, DSS*/> &array_in)
	{
		val_T **ptr;
	
		//segmented_linked_list_customsize_iterator<val_T *, DSS, FALSE> my_iter;
		sllc<T_list_elem<val_T>>::iterator_T my_iter;
		my_iter.set_list(&this->pointers_a);
	
#if 0	
		SFOREACH2
		//(
			///val_T, DSS, &array_in,
			//sllc<val_T, DSS>, 
			(&array_in)
			{
				ptr = my_iter.get_currdata();				
				**ptr = *item;	
				
				my_iter.next();					
				if (my_iter.at_end())
					break;
			}		
		//);
#else

		bool at_end;										
		sllc<val_T, DSS>::iterator_T iter(&array_in, &at_end);			
		sllc<val_T, DSS>::elem_T *item;								
															
		while(!at_end)										
		{													
			item = iter.get_currdata();
									
			{
				ptr = my_iter.get_currdata();				
				**ptr = *item;	
				
				my_iter.next();					
				if (my_iter.at_end())
					break;
			}
													
			iter.next();									
			at_end = iter.at_end_notempty();				
		};													

#endif		
	};

};

///#######################################################################

typedef my_ptr_list<A>														flist;

///#######################################################################
///
///#######################################################################

#if 0
class if_pair {

public:
	int id;
	A val;

	if_pair(int id_in, A val_in)
	{
		id = id_in;
		val = val_in;
	};				 
};
#endif

///#######################################################################
///
///#######################################################################

 ///TODO: use http://www.boost.org/libs/tuple/doc/tuple_users_guide.html#tiers instead
 

#if 0

class sv_cvpair {

public:
	int id;
	cell_value	val;

	sv_cvpair(int id_in = -1, A val_in = 1, A importance_in = 1)
	{
		id = id_in;
		
		val.set(val_in, importance_in);
	};				 
};

//#######################################################################

template 
< 
	S SS = DSS
>
class sv_cvpair_a : public sllc<sv_cvpair, SS> {};

#else

class sv_cvpair : public pair<int, cell_value> {
#if 1
public:
	sv_cvpair(int id_in = -1, A val_in = 1, A importance_in = 1)
	{
		first = id_in;
		
		second.set(val_in, importance_in);
	};
#endif	
};

///#######################################################################

class sv_cvpair_a : public map<int, cell_value> {};

#endif

///#######################################################################
///
///#######################################################################

class status_data {

public:
	int					status_pin_index;
	sv_cvpair_a			vals;
	A					activity;							///used for ACTION_STRENGTH_MEANS_ACTION_DURATION


	status_data(IN int pin_index_in, IN A activity_in = 0)		
	{ 
		status_pin_index	= pin_index_in; 
		activity			= activity_in;
	};

};

///#######################################################################
///
///#######################################################################

template
<
	typename net_T
>
class status_vector {

public:

	typedef status_vector				my_T;
	typedef map<int, status_data>		map_T;

	net_T						*net;
	map_T						status_datas;
	int							num_statuses;

	///##################################################################

protected:

	status_data *get_status_data(int status_id)
	{		
		map<int, status_data>::iterator iter;
		iter = status_datas.find(status_id);
		
		if (iter == status_datas.end())
		{
			return NULL;
		}
		else
		{
			return &iter->second;
		}
	};

	///##################################################################

	C status_data *get_status_data_const(int status_id) C
	{		
		map<int, status_data>::const_iterator iter;
		iter = status_datas.find(status_id);
		
		if (iter == status_datas.end())
		{
			return NULL;
		}
		else
		{
			return &iter->second;
		}
	};

	///##################################################################

public:

	int get_status_index_const(int status_id) C
	{
		C status_data *data = this->get_status_data_const(status_id);
		
		if (data == NULL)
		{
			return -1;
		}
		else
		{
			return data->status_pin_index;
		}
	};

	///##################################################################

	int get_status_index(int status_id, bool insert_if_not_found = FALSE)
	{
		C status_data *data = this->get_status_data_const(status_id);
		
		if (data == NULL)
		{
			if (insert_if_not_found)
			{
				status_datas.insert(pair<int, status_data>(status_id, status_data(num_statuses, 0)));
				int newindex = num_statuses;
				net->set_output_atindex(newindex, TA(0));				///clean up old unused crap
				
				num_statuses++;
				
				return newindex;
			}
			else
			{
				return -1;
			}
		}
		else
		{
			return data->status_pin_index;
		}
	};
	
	///##################################################################

	A get_status_activity_local(int status_id, bool insert_if_not_found = FALSE)
	{
		C status_data *data = this->get_status_data_const(status_id);
		
		if (data == NULL)
		{
			if (insert_if_not_found)
			{
				pair<map_T::const_iterator, bool> result = status_datas.insert(pair<int, status_data>(status_id, status_data(num_statuses, 0)));
				//int newindex = num_statuses;
				//net->set_output_atindex(newindex, TA(0));				///clean up old unused crap
				
				num_statuses++;
				
				//return newindex;
				map_T::const_iterator &iter = result.first;
				status_data C &added_data = iter->second;
				return added_data.activity;
			}
			else
			{
				//return -1;
				return 0;
			}
		}
		else
		{
			return data->activity;
		}
	};
	
	///##################################################################

	V set_status_activity_local(int status_id, A val_in, bool insert_if_not_found)
	{
		status_data *data = this->get_status_data(status_id);
		
		if (data == NULL)
		{
			if (insert_if_not_found)
			{
				pair<map_T::iterator, bool> result = status_datas.insert(pair<int, status_data>(status_id, status_data(num_statuses, 0)));
				//int newindex = num_statuses;
				//net->set_output_atindex(newindex, TA(0));				///clean up old unused crap
				
				num_statuses++;
				
				//return newindex;
				map_T::iterator &iter = result.first;
				status_data &added_data = iter->second;
				added_data.activity = val_in;
			}
			else
			{
				//return -1;
				//return 0;
			}
		}
		else
		{
			data->activity = val_in;
		}
	};
	
	///##################################################################

	status_vector()
	{
		num_statuses = 0;
	};

	///##################################################################

	V set_net(IN net_T *net_in)
	{
		net = net_in;	
	};

	///##################################################################

	V enum_statuses(int first_id, ...)		///for init		 
	{
		int id = first_id;
		///int index;
		va_list marker;

		va_start(marker, first_id);							///TODO: use "variable argument lists -> How to: Accept Variable Arguments"
		while (id != -1)			   ///-1 is terminator
		{
			/*index = */this->get_status_index(id, TRUE);
			/*if (index == -1)
			{					
				status_datas.insert(pair<int, status_data>(id, status_data(num_statuses)));
				num_statuses++;
			}*/
			
			id = va_arg(marker, int);
		}
		va_end(marker);	
	};

	///##################################################################
	///
	///##################################################################
	
	V reset_status_values() 
	{
		for (int i = 0; i < num_statuses; i++)
		{
			net->set_output_atindex(i, TA(0));
		}
	};

	///##################################################################
	
	V operator = (int status_id)
	{ 	
		this->activate_status(status_id);
	};

	///##################################################################
	
	V reset_status_values2() 
	{
		set_status_activity_local(TURN_LEFT, 0, FALSE);
		set_status_activity_local(STRAIGHT, 0, FALSE);
		set_status_activity_local(TURN_RIGHT, 0, FALSE);
	};

	///##################################################################
	
	template <typename output_data_delayer_T>
	A activate_status (int status_id, A val_arg = 1, bool read_only = false, output_data_delayer_T *output_data_delayer = NULL) const
	{ 		
		MYASSERT(read_only);
		
		return const_cast<my_T *>(this)->activate_status(status_id, val_arg, true, output_data_delayer);		
	};
	
	///##################################################################
	
	V set_output(A val_to_net)
	{
	
		if (NUM_OUTPUTS == 1)
		{
			net->set_output_atindex(0, val_to_net);			
		}
		else if (NUM_OUTPUTS == 2)
		{
			if (val_to_net > 0)
			{
				net->set_output_atindex(0, val_to_net);	
				net->set_output_atindex(1, 0.0f);	
			}
			else
			{
				net->set_output_atindex(0, 0.0f);	
				net->set_output_atindex(1, -val_to_net);
			}
		}
		else
		{
			MYASSERT(false);
		}
			
	};	
	
	///##################################################################
	
	A get_output() C
	{
	
		if (NUM_OUTPUTS == 1)
		{
			return net->get_output_atindex(0);			
		}
		else if (NUM_OUTPUTS == 2)
		{
			A v1 = net->get_output_atindex(0);
			A v2 = net->get_output_atindex(1);
			
			if (v1 > v2)
				return v1;
			else	
				return -v2;				
		}
		else
		{
			MYASSERT(false);
			
			return 0;
		}
			
	};	
	
	///##################################################################
	
	template <typename output_data_delayer_T>
	A activate_status (int status_id, A val_arg = 1, bool read_only = false, output_data_delayer_T *output_data_delayer = NULL)
	{ 		
	
	#if !ACTION_STRENGTH_MEANS_ACTION_DURATION
	
	
		int status_index = this->get_status_index(status_id, TRUE);
		
		if (status_index != -1)
		{
			if (output_data_delayer == NULL)
			{
				net->set_output_atindex(status_index, val_arg);			   ///TODO ?!
			}
			else
			{
				IFC_S(!boost::is_void<output_data_delayer_T>::value)
				{
					output_data_delayer->set(status_index, status_pair(status_id, val_arg));													 ///TODO ?!
					net->set_output_atindex(status_index, output_data_delayer->get()[status_index].val);		///TODO ?!
				}
			}
			
			//net->outputs_tick_echolines_first_cell();
			net->tick_echolines_first_cell();		///must tick all echolines because during manual control s.a is suppressed and inputs too are yet unticked

		}
		
		
	#else
	
	
		#if 1
		
			A val = get_status_activity_local(status_id, TRUE);
			
			IFC (1 || (val == 0))			///this status id was not active during previous timestep. Ergo new status_id is activated. Lets reset all time-counters.
			{	
				//set_status_activity_local(STRAIGHT, 0, FALSE);
				val += reciprocal(duration_one_timesteps_quota); 
				
				
				IFC (0 || (status_id != STRAIGHT))									///ignore straight s.a tries
				{
					A left_val = get_status_activity_local(TURN_LEFT, TRUE);
					A straight_val = get_status_activity_local(STRAIGHT, TRUE);
					A right_val = get_status_activity_local(TURN_RIGHT, TRUE);
				
					if (status_id == TURN_RIGHT)
					{
						left_val -= 1 * val;
						left_val = maxf(left_val, 0);
						
						straight_val -= 1 * val;
						straight_val = maxf(straight_val, 0);
						
						//val = right_val;
						right_val = val;
						right_val = minf(right_val, 0.75);
					}
					else if (status_id == TURN_LEFT)
					{
						right_val -= 1 * val;
						right_val = maxf(right_val, 0);
						
						straight_val -= 1 * val;
						straight_val = maxf(straight_val, 0);
						
						//val = left_val;
						left_val = val;
						left_val = minf(left_val, 0.75);
					}
				
				
					if (!read_only)
					{
						set_status_activity_local(TURN_LEFT, left_val, FALSE);
						set_status_activity_local(STRAIGHT, straight_val, FALSE);
						set_status_activity_local(TURN_RIGHT, right_val, FALSE);
					}
				}
				else
				{
					A left_val = get_status_activity_local(TURN_LEFT, TRUE);
					A right_val = get_status_activity_local(TURN_RIGHT, TRUE);
					
					left_val -= 1 * val;
					left_val = maxf(left_val, 0);
					
					right_val -= 1 * val;
					right_val = maxf(right_val, 0);
					
					A straight_val = val;
					straight_val = minf(straight_val, 0.75);
					
					
					if (!read_only)
					{
						set_status_activity_local(TURN_LEFT, left_val, FALSE);
						set_status_activity_local(STRAIGHT, straight_val, FALSE);
						set_status_activity_local(TURN_RIGHT, right_val, FALSE);
					}
				}
			}
			else
			{
				//if (status_id != STRAIGHT)
				//{
				val += reciprocal(duration_one_timesteps_quota); 
				
				
				if (!read_only)
				{
					set_status_activity_local(status_id, val, FALSE);
				}
				
				//}
			}
			
			
			
			A val_out = 0;
			A val_to_net = 0;
			//statuses status_out = STOP;
			
			switch (status_id)
			{
				case STRAIGHT:
				{
					val_out = 0;
					val_to_net = val_out;
					/*
					val_out = get_status_activity_local(TURN_LEFT, FALSE);				///lets take cached value
					if (val_out > 0)
					{
						val_out = -val_out;
					}
					else
					{
						val_out = get_status_activity_local(TURN_RIGHT, FALSE);
					}
					*/
				}
				break;
				
				case TURN_LEFT:
				{
					val_out = -val - 0.25;
					
					val_out = maxf(val_out, -1.00);
					
					
					val_to_net = val_out; 					
					if (use_action_duration != 0)
					{
						val_to_net -= inc_action_duration_sent_to_net_by;
					}					
					
				}
				break;
				
				case TURN_RIGHT:
				{
					val_out = val + 0.25;
					
					val_out = minf(val_out, 1.00);
					

					val_to_net = val_out; 					
					if (use_action_duration != 0)
					{
						val_to_net += inc_action_duration_sent_to_net_by;
					}
					
				}
				break;
			};
				
			
			if (!read_only)
			{	
				//net->set_output_atindex(0, val_out);	
				
				if (output_data_delayer == NULL)
				{
					set_output(val_to_net);
				}
				else
				{
					IFC_S(!boost::is_void<output_data_delayer_T>::value)
					{
						output_data_delayer->set(0, status_pair(status_id, val_to_net));												
						//net->set_output_atindex(0, output_data_delayer->get()[0].val);		
						set_output(output_data_delayer->get()[0].val);
					}
				}
				
				//net->outputs_tick_echolines_first_cell();
				net->tick_echolines_first_cell();		///must tick all echolines because during manual control s.a is suppressed and inputs too are yet unticked
			}
			
			
			return val_out;
		
		#else
		
			//A val = get_status_activity_local(status_id, TRUE);
			//A val = net->get_output_atindex(0);
			A val = get_output();
			
			switch (status_id)
			{
				case STRAIGHT:
				{
					//val = 0;
				}
				break;
				
				case TURN_LEFT:
				{
					val -= reciprocal(duration_one_timesteps_quota);
					val = maxf(val, -1.00);
				}
				break;
				
				case TURN_RIGHT:
				{
					val += reciprocal(duration_one_timesteps_quota);
					val = minf(val, 1.00);
				}
				break;
			};
			
			//net->set_output_atindex(0, val);	
			if (output_data_delayer == NULL)
			{
				//net->set_output_atindex(0, val);				
				set_output(val);
			}
			else
			{
				IFC_S(!boost::is_void<output_data_delayer_T>::value)
				{
					output_data_delayer->set(0, status_pair(status_id, val));											
					//net->set_output_atindex(0, output_data_delayer->get()[0].val);		
					set_output(output_data_delayer->get()[0].val);
				}
			}
			
			//net->outputs_tick_echolines_first_cell();
			net->tick_echolines_first_cell();		///must tick all echolines because during manual control s.a is suppressed and inputs too are yet unticked
			
		
		#endif
		
			
	#endif
	};

	///##################################################################
#if 0
	V pass_statuses_to_net_outputs()
	{ 	
		///done already
	};

	///##################################################################
	
	A read_statuses_from_net_outputs() C 
	{
		
	};
#endif
	///##################################################################
	
	A get_status_val(int status_id, bool insert_if_not_found = FALSE) 
	{
		int status_index = this->get_status_index(status_id, insert_if_not_found);
		
		if (status_index != -1)
		{
			return net->get_output_atindex(status_index);	
		}
		else
		{
			return NAN;	
		}
	};
	
	///##################################################################
	
	cell_value get_status_cv(int status_id, bool insert_if_not_found = FALSE) 
	{
		int status_index = this->get_status_index(status_id, insert_if_not_found);
		
		if (status_index != -1)
		{
			return net->get_output_cv_atindex(status_index);	
		}
		else
		{
			return cell_value(NAN, NAN);	
		}
	};
	
	///##################################################################
	
	cell_value get_status_predicted_diff_cv(int status_id, bool insert_if_not_found = FALSE) 
	{
		int status_index = this->get_status_index(status_id, insert_if_not_found);
		
		if (status_index != -1)
		{
			return net->get_output_cv_predicted_diff_atindex(status_index);	
		}
		else
		{
			return cell_value(NAN, NAN);	
		}
	};
		
	///##################################################################
	
	V set_status_params_for(IN int status_id, IN sv_cvpair param1_id_val, ...)
	{
		this->enum_statuses(status_id, -1);
		
	
		status_data *data = this->get_status_data(status_id);
		
		
		
		sv_cvpair val = param1_id_val;
		va_list marker;

		va_start(marker, param1_id_val);			  ///TODO: use "variable argument lists -> How to: Accept Variable Arguments"
		while (val.first != -1)			   ///-1 is terminator
		{
			map<int, cell_value>::iterator iter;	
		
		
			iter = data->vals.find(val.first);
			
			if (iter == data->vals.end())
			{
				data->vals.insert(pair<int, cell_value>(val.first, val.second));
			}
			else
			{
				iter->second = val.second;
			}
			
			
			val = va_arg(marker, sv_cvpair);
		}
		va_end(marker);			
		
	};
	
	///##################################################################
	
	V get_blended_status_params(INOUT float_a *param_vals, IN int param1_id, ...) C
	{
		int id = param1_id;
		
		va_list marker;
		va_start(marker, param1_id);		///TODO: use "variable argument lists -> How to: Accept Variable Arguments"
		while (id != -1)			   ///-1 is terminator
		{
			///lets search for current param's usage in every status block and compute the mean value
		

			cell_value vals_sum;
			
			cell_value status_active;
			cell_value param_val;
			
			
			///lets walk the statuses array
			map<int, status_data>::const_iterator status_iter;
			status_iter = status_datas.begin();
			while (status_iter != status_datas.end())
			{
				C status_data *params = &status_iter->second;
				
				map<int, cell_value>::const_iterator param_iter;
				param_iter = params->vals.find(id);
				if (param_iter != params->vals.end())					///param data found in status
				{
					int pin_index = params->status_pin_index;
					status_active = net->get_output_cv_atindex(pin_index);
					param_val = param_iter->second;
					
					///lets compute current statuses contribution by the product of its activation and its value/importance
					vals_sum = cvop::add_for_status_vector(vals_sum, cvop::mul_for_status_vector(param_val, status_active));	
				}
				
				status_iter++;	   
			}
			
			A val_out = vals_sum.mean();
			param_vals->additem(val_out);
					
		
			id = va_arg(marker, int);
		}		
		va_end(marker);	
		
	};

};

///#######################################################################
///
///#######################################################################



}	///namespace ai




#endif		///#ifndef STATUS_VECTOR_H
